#include "stdafx.h"
#include "AdaTreeNatWrapper.h"
#include "clix.h"

using namespace clix;
using namespace AdaTreeNat;
using namespace System;
using namespace System::Collections;
using namespace XmlClassifier::DataStructures;

AdaTreeNatWrapper::AdaTreeNatWrapper(bool IsOrderedTree, bool IsTopDownSubtree, bool IsIncremental,
									 bool IsSlidingWindow, bool IsAdaptive1, bool IsAdaptive2, bool IsRelaxedSupport,
									 bool IsLogRelaxedSupport, int BatchSize, int SlidingWindowSize, int NumberRelaxedIntervals, double RelativeMinSup, bool OnlyMaximal)
{
	this->IsOrderedTree = IsOrderedTree;
	this->IsTopDownSubtree = IsTopDownSubtree;
	this->IsIncremental = IsIncremental; 
	this->IsLabelled = IsLabelled;
	this->IsSlidingWindow = IsSlidingWindow; 
	this->IsAdaptive1 = IsAdaptive1;
	this->IsAdaptive2 = IsAdaptive2;
	this->IsRelaxedSupport = IsRelaxedSupport;
	this->IsLogRelaxedSupport = IsLogRelaxedSupport; 
	this->BatchSize = BatchSize; 
	this->SlidingWindowSize = SlidingWindowSize; 
	this->NumberRelaxedIntervals = NumberRelaxedIntervals; 
	this->RelativeMinSup = RelativeMinSup;
	this->BatchNum = 0;
	this->OnlyMaximal = OnlyMaximal;

#if defined(LABELLED)
	this->IsLabelled = true;
#else
	this->IsLabelled = false;
#endif
	this->IsTreeNat = (!this->IsIncremental && !this->IsSlidingWindow &&!this->IsAdaptive1 && !this->IsAdaptive2);

	if (this->IsOrderedTree==true){
		if (this->IsTopDownSubtree==true)
			this->TFactory = new OrderedTree_TopDown_Factory;
		else
			this->TFactory = new OrderedTreeFactory;
	} else {
		if (this->IsTopDownSubtree==true)
			this->TFactory = new UnorderedTree__TopDown_Factory;
		else
			this->TFactory = new UnorderedTreeFactory;
	}
}

AdaTreeNatWrapper::~AdaTreeNatWrapper()
{
	this->DataReader->close();
	delete this->DataReader;
}

void AdaTreeNatWrapper::MarshalList(Generic::List<String^>^ list)
{
	if (list == nullptr) {
		throw gcnew ArgumentNullException(L"list");
	}
	if(this->BatchDocuments != NULL) {
		this->BatchDocuments->clear();
	} else {
		this->BatchDocuments = new std::vector<string>();
	}

	this->BatchDocuments->reserve(list->Count);
	for(int i=0; i< list->Count; i++)
		this->BatchDocuments->push_back(marshalString<E_ANSI>(list[i]));
}

Generic::List<Pattern^>^ AdaTreeNatWrapper::GetPatterns(Generic::List<String^>^ batch, int batchClass)
{
	Generic::List<Pattern^>^ patterns = gcnew Generic::List<Pattern^>();
	this->MarshalList(batch);

	//// Wskaniki zarzdzane na natywne
	this->BatchSize = this->BatchDocuments->size();
	int bSize=this->BatchSize;
	mBatchSize=&bSize;
	int WindowSize=this->WindowSize;
	mWindowSize=&WindowSize;

	if(this->BatchNum == 0)
	{
		WindowSize = bSize;

		if (this->IsTreeNat == false) {
			if (this->RelativeMinSup > 1) {
				error(" Support must be a number between 0 and 1.");
			}

			this->DataReader = new DatasetReader(this->BatchDocuments, this->IsOrderedTree, this->IsLabelled, this->RelativeMinSup, this->BatchSize);

			if (this->RelativeMinSup < 1) {
				this->MinSup = (int) (this->RelativeMinSup * this->DataReader->size());
			}
			if (this->MinSup==0) {
				this->MinSup=1;
			}
		} else if (this->IsTreeNat == true) {

			this->DataReader = new DatasetReader(this->BatchDocuments,this->IsOrderedTree, this->IsLabelled,this->RelativeMinSup,0);

			if (this->RelativeMinSup < 1) {
				this->MinSup= (int) (this->RelativeMinSup * this->DataReader->size());
			}
			if (this->MinSup == 0) {
				error(" Support can not be zero.");
			}
		}

		mdataset_reader= this->DataReader;

		mdblmin_sup=this->RelativeMinSup;
		mtree_factory=this->TFactory;
		if (this->IsRelaxedSupport==true){
			if (this->IsLogRelaxedSupport==true)
				support_factory = new Log_Relaxed_Support_Factory(this->NumberRelaxedIntervals, mdblmin_sup);
			else
				support_factory = new Linear_Relaxed_Support_Factory(this->NumberRelaxedIntervals, mdblmin_sup);
		} else {
			support_factory = new Linear_Relaxed_Support_Factory(this->SlidingWindowSize, mdblmin_sup);
		}
		mclosedTrees = new ClosedTrees(support_factory, this->TFactory, this->IsAdaptive2, mBatchSize, mWindowSize, mdblmin_sup);



		//AdaTreeNat Algorithm
		this->SlidingWindow = new list<TreeDataSet>();
		#if defined(LABELLED)
		this->SlidingWindowAux= new list<TreeNodesAux>();
		#endif
		if (this->IsSlidingWindow==true){
			this->SlidingWindow->push_back(this->DataReader->get_dataset());
#if defined(LABELLED)
			this->SlidingWindowAux->push_back(this->DataReader->get_dataset_aux());;
#endif
		}

		//Mining first batch of data
		TreeMining *closed_tree_mining = new TreeMining(this->TFactory, this->DataReader, this->MinSup);
		mapTree *ClosedMapTree= closed_tree_mining->get_ClosedMapTree();
		for( mapTreeIterator iter = ClosedMapTree->begin(); iter != ClosedMapTree->end(); ++iter ) {
			mclosedTrees->insert(iter->second->get_TreeNodes(),support_factory->RelaxedSupport(iter->first,*mBatchSize),false);
		}
		mclosedTrees->EraseNotClosedTrees(); // Cleans not closed items due to new relaxed support
		mclosedTrees->TagNotMaximalTrees();
		//mclosedTrees->print();

		//Init Lattice
		closed_tree_mining->Compress_MapTree();
		delete closed_tree_mining;


		AdNumberClosedTrees = new Adwin(5);
		if (this->IsAdaptive1 == true){
			AdNumberClosedTrees->update(mclosedTrees->NumberClosedTrees());
			this->SlidingWindowSize=2*BatchSize;
		}

	} else {

		if (this->IsSlidingWindow==true){
			//Delete transactions

			if (WindowSize>=this->SlidingWindowSize){
				for (int i=WindowSize-this->SlidingWindowSize;i>=0;i-=BatchSize){
					WindowSize-=BatchSize;
					TreeDataSet D = this->SlidingWindow->front();
					this->SlidingWindow->pop_front();
#if defined(LABELLED)
					TreeNodesAux Daux = this->SlidingWindowAux->front();
					this->SlidingWindowAux->pop_front();
#endif
					this->MinSup = (int) (this->RelativeMinSup*BatchSize);
					if (this->MinSup<1) this->MinSup=1;
#if defined(LABELLED)
					Delete(&D,&Daux,this->MinSup);
#else
					Delete(&D,min_sup);
#endif
				}
			}
		}

		//Add transactions
		this->DataReader->read_Next_Batch(this->BatchDocuments);
		*mBatchSize=this->DataReader->size();
		if (this->DataReader->size()==0) return patterns;
			;
		this->MinSup = (int) (this->RelativeMinSup*this->DataReader->size());
		if (this->MinSup < 1) this->MinSup = 1;

		//Add treedataset to Sliding Window
		if (this->IsSlidingWindow==true ){
			this->SlidingWindow->push_back(this->DataReader->get_dataset());
#if defined(LABELLED)
			this->SlidingWindowAux->push_back(this->DataReader->get_dataset_aux());;
#endif
		}
		WindowSize+=*mBatchSize;


		TreeMining *closed_tree_mining= new TreeMining(this->TFactory, this->DataReader, this->MinSup);
		closed_tree_mining->Compress_MapTree();
		Add(closed_tree_mining->get_ClosedMapTree());
		delete closed_tree_mining;

		if ( this->IsAdaptive2==false ){
			mclosedTrees->EraseNotClosedTrees();
		}

		//ADWIN1
		if (this->IsAdaptive1==true ){
			if (this->AdNumberClosedTrees->update(mclosedTrees->NumberClosedTrees()))
				this->SlidingWindowSize=this->AdNumberClosedTrees->length()*BatchSize;
			else
				this->SlidingWindowSize+=BatchSize;
		}

		//ADWIN2
		if ( this->IsAdaptive2 == true ){
			if (mclosedTrees->get_Change()>1) {
				mclosedTrees->set_Change(0);
				mclosedTrees->EraseNotClosedTrees();
			}
		}
		mclosedTrees->TagNotMaximalTrees();
		//mclosedTrees->print();
	}

	
	this->BatchNum++;
	//// Wskaniki natywne do zarzdzanych
	this->BatchSize = *mBatchSize;
	this->WindowSize= *mWindowSize;


	for( mapTreeIterator iter = mclosedTrees->getPatterns()->begin(); iter != mclosedTrees->getPatterns()->end(); ++iter ) {	
		//// Drzewo musi mie odpowiedni support
		if (!mclosedTrees->isInsufficientRelaxedSupport(iter->second->get_Support())
			///i (w zalenoci od parametru) by maksymalne
				&& (!this->OnlyMaximal || !(iter->second)->get_isNotMaximal())){
				
					Pattern^ p = gcnew Pattern();
					p->Support = (iter->second)->get_Support();
					p->Size = (iter->second)->get_Size();
					p->EncodedPattern = gcnew String((iter->second)->zaki_string().c_str());
					p->Class = batchClass;
					patterns->Add(p);
		}
	}

	return patterns;
}

////////////////////////////////////////////////////////////////////////////////

void AdaTreeNatWrapper::Add(mapTree *ClosedMapTree){
	list<TreeNodes*> list_Insert_TreeNodes;
	list<int> list_Insert_Support;
	TreeDataSet *D;
	SupportContainer *SupportTreesDataSet =new SupportContainer;
	D = MapTreeToDataset(ClosedMapTree,SupportTreesDataSet);

#if defined(LABELLED)
	for(Labels_iterator label=mdataset_reader->get_labels()->begin(); label != mdataset_reader->get_labels()->end(); ++label){
		Tree*t =mtree_factory->create_tree(D,*label);
		//t->isNotMaximal = 
		if (t->get_Support()==0) break;
		t->parentsLabel.push_back(*label);
#else
	Tree *t= mtree_factory->create_tree(D);
#endif
	int SupportT2=SupportTreesDataSet->at(t->Id_Front());

	// if it is In Lattice, updateSupport otherwise insert it
	if (mclosedTrees->updateSupportAdd(t->get_TreeNodes_(), SupportT2)==false) {
		int closure_support=0;
		TreeNodes treeclosure=mclosedTrees->closure(t,&closure_support);
		int AbsSupport= (int) (support_factory->RelativeSupport(closure_support)*(*mWindowSize-*mBatchSize)+SupportT2);
		if ( AbsSupport>mdblmin_sup* (*mWindowSize)){
			if (treeclosure.size()<=1 && t->get_Size()==D->at(t->Id_Front()).size()) {
				//There is no supertree with the same support
				list_Insert_TreeNodes.push_back(t->get_TreeNodes());
				list_Insert_Support.push_back(support_factory->RelaxedSupport(SupportT2,*mWindowSize));
			}
			else if (treesIntersection(treeclosure,D->at(t->Id_Front())) == t->get_TreeNodes_()){
				//Insert a new closed tree into Lattice
				list_Insert_TreeNodes.push_back(t->get_TreeNodes());
				list_Insert_Support.push_back(support_factory->RelaxedSupport(AbsSupport,*mWindowSize));

			}
		}
	}

	recursive_Add(t,D,SupportTreesDataSet,&list_Insert_TreeNodes,&list_Insert_Support);
#if defined(LABELLED)
}
#endif

//Add new closed trees
list<int>::iterator it_sup=list_Insert_Support.begin();
for (list<TreeNodes*>::iterator it=list_Insert_TreeNodes.begin();it!=list_Insert_TreeNodes.end();it++){
	mclosedTrees->insert(*it, *it_sup);
	it_sup++;
}
mclosedTrees->UpdateSupport(ADD_OPTION);
}

////////////////////////////////////////////////////////////////////////////////

void AdaTreeNatWrapper::recursive_Add(Tree *t,TreeDataSet *D,SupportContainer *SupportDataset, list<TreeNodes*> *list_Insert_TreeNodes, list<int> *list_Insert_Support){
	Integer len_tail=t->get_lenTail();
	//For  every tt  that can be extended from t in one step
#if defined(LABELLED)
	for (int depth=len_tail+1;depth>0;depth--){
		Labels *labels;
		int level=len_tail+1-depth;
		if (depth==0)
			labels=mdataset_reader->get_labels();
		else if (depth==1)
			labels=mdataset_reader->get_labels(t->currentAncestorLabel(level));
		else if (depth==2)
			labels=mdataset_reader->get_labels(t->currentAncestorLabel(level),t->currentAncestorLabel(level+1));
		else
			labels=mdataset_reader->get_labels(t->currentAncestorLabel(level),t->currentAncestorLabel(level+1),t->currentAncestorLabel(level+2));
		for(Labels_iterator label=labels->begin(); label != labels->end(); ++label){
			Tree *tt = mtree_factory->create_tree(t);
			Node extension(depth,*label);

#else
	for (int extension=len_tail+1;extension>0;extension--){
		Tree *tt = mtree_factory->create_tree(t);
#endif
		tt->extend(extension);
		if (tt->isCanonicalRepresentative(extension)) {
			tt->supportInc(D,1);
#if defined(LABELLED)
			tt->set_lenTail(depth);
			if (tt->parentsLabel.size()==depth)
				tt->parentsLabel.push_back(*label);
			else
				tt->parentsLabel.at(depth)=*label;
#else
			tt->set_lenTail(extension);
#endif

			if (tt->get_Support()!=0) {
				int SupportT2=SupportDataset->at(tt->Id_Front());

				if (mclosedTrees->updateSupportAdd(tt->get_TreeNodes_(), SupportT2)==false) {
					//Check closure
					int closure_support=0;
					TreeNodes treeclosure=mclosedTrees->closure(tt,&closure_support);
					int AbsSupport= (int) (support_factory->RelativeSupport(closure_support)*(*mWindowSize-*mBatchSize)+SupportT2);
					if ( AbsSupport>mdblmin_sup* (*mWindowSize)){
						if (treeclosure.size()<=1 && tt->get_Size()==D->at(tt->Id_Front()).size()) {
							list_Insert_TreeNodes->push_back(tt->get_TreeNodes());
							list_Insert_Support->push_back(support_factory->RelaxedSupport(SupportT2,*mWindowSize));
						}
						else if (treeclosure.size()>0 &&treesIntersection(treeclosure,D->at(tt->Id_Front())) == tt->get_TreeNodes_()){

							list_Insert_TreeNodes->push_back(tt->get_TreeNodes());
							list_Insert_Support->push_back(support_factory->RelaxedSupport(AbsSupport,*mWindowSize));
						}
					}
				}
				recursive_Add(tt,D, SupportDataset,list_Insert_TreeNodes, list_Insert_Support);
			}
		}
	}
#if defined(LABELLED)
}
#endif
}

////////////////////////////////////////////////////////////////////////////////

#if defined(LABELLED)
void AdaTreeNatWrapper::Delete(TreeDataSet *D,TreeNodesAux *Daux, int min_sup){
#else
void AdaTreeNat_Mining::Delete(TreeDataSet *D,int min_sup){
#endif
	SupportContainer *SupportTreesDataSet =new SupportContainer;
	listTree *TreesCheckClosureList= new listTree;
#if defined(LABELLED)
	mDatasetAux=Daux;
	for(Labels_iterator label=mdataset_reader->get_labels()->begin(); label != mdataset_reader->get_labels()->end(); ++label){
		Tree* t=mtree_factory->create_tree(D,*label);
		if (t->get_Support()==0) break;
		t->parentsLabel.push_back(*label);
#else
	Tree* t=mtree_factory->create_tree(D);
#endif


	mclosedTrees->updateSupportDelete(t->get_TreeNodes_(), t->get_Support(),TreesCheckClosureList);
	recursive_Delete(t,D,min_sup,SupportTreesDataSet, TreesCheckClosureList);

	delete t;
#if defined(LABELLED)
}
#endif

mclosedTrees->UpdateSupport(DELETE_OPTION);
mclosedTrees->Check_Closure_List_Delete(TreesCheckClosureList);
delete TreesCheckClosureList;

}

////////////////////////////////////////////////////////////////////////////////

void AdaTreeNatWrapper::recursive_Delete(Tree *t,TreeDataSet *D,int min_sup,SupportContainer *SupportDataset, listTree *TreesCheckClosureList){
	Integer len_tail=t->get_lenTail();
	//For  every tt  that can be extended from t in one step
#if defined(LABELLED)
	for (int depth=len_tail+1;depth>0;depth--){
		Labels *labels;
		int level=len_tail+1-depth;
		if (depth==0)
			labels=mDatasetAux->get_labels();
		else if (depth==1)
			labels=mDatasetAux->get_labels(t->currentAncestorLabel(level));
		else if (depth==2)
			labels=mDatasetAux->get_labels(t->currentAncestorLabel(level),t->currentAncestorLabel(level+1));
		else
			labels=mDatasetAux->get_labels(t->currentAncestorLabel(level),t->currentAncestorLabel(level+1),t->currentAncestorLabel(level+2));
		for(Labels_iterator label=labels->begin(); label != labels->end(); ++label){
			Tree *tt = mtree_factory->create_tree(t);
			Node extension(depth,*label);
#else
	for (int extension=len_tail+1;extension>0;extension--){
		Tree *tt = mtree_factory->create_tree(t);
#endif
		tt->extend(extension);
		if (tt->isCanonicalRepresentative(extension)) {
			tt->supportInc(D,min_sup);
#if defined(LABELLED)
			tt->set_lenTail(depth);
			if (tt->parentsLabel.size()==depth)
				tt->parentsLabel.push_back(*label);
			else
				tt->parentsLabel.at(depth)=*label;
#else
			tt->set_lenTail(extension);
#endif
			if (tt->get_Support()!=0) {
				mclosedTrees->updateSupportDelete(tt->get_TreeNodes_(), tt->get_Support(),TreesCheckClosureList);
				recursive_Delete(tt,D,min_sup, SupportDataset,TreesCheckClosureList);
			}
		}
	}
#if defined(LABELLED)
}
#endif
}

////////////////////////////////////////////////////////////////////////////////

TreeDataSet* AdaTreeNatWrapper::MapTreeToDataset(mapTree *ClosedMapTree, SupportContainer *SupportDataset) {
	//MapTreeToDataset: Gets a TreeDataset from a mapTree
	TreeDataSet *D=new TreeDataSet;

	for( mapTreeRevIterator iter = ClosedMapTree->rbegin(); iter != ClosedMapTree->rend(); ++iter ) {
		TreeNodes tn=(iter->second)->get_TreeNodes_();
		int Support=iter->first;
		D->push_back(tn);
		SupportDataset->push_back(Support);
	}
	return D;
}

////////////////////////////////////////////////////////////////////////////////

TreeNodes AdaTreeNatWrapper::treesIntersection(TreeNodes t1, TreeNodes t2){
	//Intersection: returns the intersection of two trees
	TreeDataSet *D=new TreeDataSet;

	D->push_back(t1);
	D->push_back(t2);

	DatasetReader* dataset_reader= new DatasetReader(D);
	TreeMining *closed_tree_mining= new TreeMining(mtree_factory, dataset_reader, 2);

	TreeNodes ret;
	
	if (closed_tree_mining->get_ClosedMapTree()->begin() != closed_tree_mining->get_ClosedMapTree()->end())
	{
		ret = (closed_tree_mining->get_ClosedMapTree()->begin()->second)->get_TreeNodes_();
	}

	return ret;
}

