#pragma once

#include "Tree.h"
#include "OrderedTree.h"
#include "UnorderedTree.h"
#include "DatasetReader.h"
#include "TreeMining.h"
#include "ClosedTrees.h"
#include "AdaTreeNat_Mining.h"
#include <string>
#include<stdlib.h>
#include<iostream>
#include<list>
#include<vector>
#include<set>
#include<map>
#include<fstream>
#include <ctime>
#include "DatasetReader.h"
#include "Tree.h"
#include "ClosedTrees.h"
#include "TreeFactory.h"
#include "SupportFactory.h"
#include <list>
#include <vector>
#include <map>
#include <fstream>

using namespace System;
using namespace System::Collections;
using namespace XmlClassifier::DataStructures;

namespace AdaTreeNat{

	public ref class AdaTreeNatWrapper
	{
	public:
		/// <summary>
        /// Konstruktor
        /// </summary>
        /// <param name="IsIncremental">Okrela czy odkrywanie wzorcw ma by przyrostowe (bez mechanizmu zapominania)</param>
        /// <param name="IsSlidingWindow">Okrela czy odkrywanie wzorcw ma by przyrostowe na oknie ostatnich dokumentw (zapominanie przez stay rozmiar okna)</param>
        /// <param name="IsAdaptive1">Okrela czy odkrywanie wzorcw ma by przyrostowe z dynamicznym oknen (dynamiczne zapominanie)</param>
        /// <param name="IsAdaptive2">Okrela czy odkrywanie wzorcw ma by przyrostowe z dynamicznym oknen (dynamiczne zapominanie)</param>
        /// <param name="IsRelaxed_Support">Czy korzysta z relaxed support</param>
        /// <param name="IsLogRelaxedSupport">Czy korzysta z log relaxed support</param>
        /// <param name="BatchSize">Rozmiar pojedynczego bloku danych, ktry bdzie przekazywany do algorytmu</param>
        /// <param name="SlidingWindowSize">Rozmiar okna przesuwnego do zapominania</param>
        /// <param name="NumberRelaxedIntervals">?</param>
        /// <param name="RelativeMinSup">Prg minimalnego wsparcia jako warto midzy 0 a 1</param>
		/// <param name="RelativeMinSup">Okrela czy bra pod uwag tylko poddrzewa maksymalne.</param>
		AdaTreeNatWrapper(bool IsOrderedTree, bool IsTopDownSubtree,
			bool IsIncremental, bool IsSlidingWindow, bool IsAdaptive1,
			bool IsAdaptive2, bool IsRelaxed_Support, bool IsLogRelaxedSupport, 
			int BatchSize, int SlidingWindowSize, int NumberRelaxedIntervals, 
			double RelativeMinSup, bool OnlyMaximal);
		
		/// <summary>
		/// Metoda pobierajca czste drzewa znalezione z/po wczytaniu kolejnego zbioru dokumentw.
		/// </summary>
		/// <param name="batch">Zbir dokumentw do analizy. Metoda powinna by w ptli odpalana z kolejnymi "batchami" dokumentw a do skoczenia si caego zbioru dokumentw.</param>
		/// <param name="batchClass">Klasa dokumentw (rzeczywista etykieta)</param>
		Generic::List<Pattern^>^ GetPatterns(Generic::List<String^>^ batch, int batchClass);

		/// <summary>
		/// Desktruktor
		/// </summary>
		~AdaTreeNatWrapper();
	private:
		bool IsOrderedTree;
		bool IsTopDownSubtree;
		bool IsIncremental; 
		bool IsLabelled;
		bool IsSlidingWindow; 
		bool IsAdaptive1;
		bool IsAdaptive2;
		bool IsRelaxedSupport;
		bool IsLogRelaxedSupport; 
		int BatchSize;
		int WindowSize;
		int SlidingWindowSize; 
		int NumberRelaxedIntervals; 
		int BatchNum; 
		int MinSup;
		double RelativeMinSup;
		bool IsTreeNat;
		bool OnlyMaximal;

		TreeFactory* TFactory;
		DatasetReader *DataReader;
		Adwin *AdNumberClosedTrees;
		list<TreeDataSet> *SlidingWindow;
		#if defined(LABELLED)
		list<TreeNodesAux> *SlidingWindowAux;
		#endif
   		ClosedTrees *mclosedTrees;
    	double mdblmin_sup;
    	int *mBatchSize;
    	int *mWindowSize;
    	TreeFactory *mtree_factory;
        Support_Factory *support_factory;
        DatasetReader* mdataset_reader;
		std::vector<string>* BatchDocuments;
        #if defined(LABELLED)
        TreeNodesAux *mDatasetAux;
        #endif

		void recursive_Add(Tree *t,TreeDataSet *D,SupportContainer *SupportDataset, list<TreeNodes*> *list_Insert_TreeNodes, list<int> *list_Insert_Support);
		void recursive_Delete(Tree *t,TreeDataSet *D,int min_sup,SupportContainer *SupportDataset, listTree *TreesCheckClosureList);
		TreeDataSet* MapTreeToDataset(mapTree *ClosedMapTree, SupportContainer *SupportDataset) ;
		TreeNodes treesIntersection(TreeNodes t1, TreeNodes t2);
		void AdaTreeNatWrapper::MarshalList(Generic::List<String^>^ list);
        void Add(mapTree *ClosedMapTree);
        #if defined(LABELLED)
        void Delete(TreeDataSet *D,TreeNodesAux *Daux, int min_sup);
        #else
        void Delete(TreeDataSet *D,int min_sup);
        #endif
	};
}