﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using XmlClassifier.DataStructures;

namespace XmlClassifier.Algorithms
{
	public class Classifier
	{
		public static string Classify(IEnumerable<Pattern> model, Document item)
		{
			foreach (var pattern in model)
			{
				pattern.CalculateDistance(item);
			}

			var closestPatterns = model.Where(x => x.GetDistance() == model.Min(y => y.GetDistance())).ToList();

			Dictionary<string, double> elections = PerformVoting(closestPatterns);

			string bestClass = SelectBestClass(elections);

			item.SetClass(bestClass);

			return bestClass;
		}

		private static Dictionary<string, double> PerformVoting(List<Pattern> closestPatterns)
		{
			Dictionary<string, double> elections = new Dictionary<string, double>();

			foreach (var neighbor in closestPatterns)
			{
				if (!elections.ContainsKey(neighbor.GetClass()))
				{
					elections.Add(neighbor.GetClass(), neighbor.GetWeight());
				}
				else
				{
					elections[neighbor.GetClass()] += neighbor.GetWeight();
				}
			}

			return elections;
		}

		private static string SelectBestClass(Dictionary<string, double> elections)
		{
			double bestResult = -1;
			string bestClass = null;
			foreach (var vote in elections)
			{
				if (vote.Value > bestResult)
				{
					bestClass = vote.Key;
					bestResult = vote.Value;
				}
			}
			return bestClass;
		}
	}
}
