﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;

namespace XmlClassifier.Algorithms
{
	/// <summary>
	/// A new version of partial tree-edit distance
	/// </summary>
	public static partial class PTED
	{
		/// <summary>
		/// Partial tree-edit distance algorithm
		/// Dynamic programming version
		/// </summary>
		/// <param name="P">Pattern tree</param>
		/// <param name="T">Document tree</param>
		/// <returns>How much does P need to be modified to become a subtree of T?</returns>
		public static int Dynamic(XElement P, XElement T)
		{
			int tmp = deltaDynamic(P, T);
			int pted = tmp;

			foreach (var u in T.Descendants())
			{
				tmp = deltaDynamic(P, u);
				if (tmp < pted)
				{
					pted = tmp;
				}
			}

			foreach (var v in P.Descendants())
			{
				tmp = deltaDynamic(v, T) + (P.Descendants().Count() - v.Descendants().Count());
				if (tmp < pted)
				{
					pted = tmp;
				}
			}

			return pted;
		}

		private static int deltaDynamic(XElement p, XElement t)
		{
			int[,] singleTable = new int[p.Elements().Count() + 1, t.Elements().Count() + 1];

			for (int j = 0; j < singleTable.GetLength(1); j++)
			{
				singleTable[0, j] = 0;
			}

			for (int i = 1; i < singleTable.GetLength(0); i++)
			{
				singleTable[i, 0] = singleTable[i - 1, 0] + p.Elements().ElementAt(i - 1).DescendantNodesAndSelf().Count();
			}

			for (int i = 1; i < singleTable.GetLength(0); i++)
			{
				for (int j = 1; j < singleTable.GetLength(1); j++)
				{
					int top = singleTable[i - 1, j] + p.Elements().ElementAt(i - 1).DescendantNodesAndSelf().Count();
					int left = singleTable[i, j - 1];
					int both = singleTable[i - 1, j - 1] + deltaDynamic(p.Elements().ElementAt(i - 1), t.Elements().ElementAt(j - 1));
					singleTable[i, j] = Math.Min(Math.Min(top, left), both);
				}
			}

			return singleTable[singleTable.GetLength(0) - 1, singleTable.GetLength(1) - 1] + (p.Name.LocalName == t.Name.LocalName ? 0 : 1);
		}
	}
}
