﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using XmlClassifier.Algorithms;
using XmlClassifier.DataStructures;
using XmlClassifier.Helpers;

namespace XmlClassifier.DataStructures
{
	public class Document
	{
		/// <summary>
		/// Konstruktor
		/// </summary>
		/// <param name="doc">Dokument zapisany w formacie Zakiego. Pierwszy znak to klasa, do której należy.</param>
		public Document(string doc)
		{
			this.ExpectedClass = int.Parse(doc.Substring(0, doc.IndexOf(' ')));
			this.Content = doc.Substring(doc.IndexOf(' ') + 1, doc.Length - doc.IndexOf(' ') - 1);
			this.XDocument = this.Content.Parse(2);
		}

		/// <summary>
		/// Klasa, do której został przypisany dokument podczas fazy testowania.
		/// Dla dokumentów ze zbioru treningowego ta wartość powinna być zawsze pusta.
		/// </summary>
		public int? ActualClass { get; set; }

		/// <summary>
		/// Faktyczna klasa dokumentu.
		/// Wartość powinna być zawsze wypełniona.
		/// </summary>
		public int ExpectedClass { get; set; }

		/// <summary>
		/// Dokument zapisany w formacie Zakiego.
		/// </summary>
		public string Content { get; set; }

		/// <summary>
		/// Dokument zapisany w formacie XML
		/// </summary>
		public XElement XDocument { get; set; }

		/// <summary>
		/// Sets the class of the document given by the classification model.
		/// </summary>
		/// <param name="c">Predicted document class</param>
		public void SetClass(string c)
		{
			this.ActualClass = int.Parse(c);
		}
	}
}
