﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using XmlClassifier.Algorithms;
using XmlClassifier.Helpers;

namespace XmlClassifier.DataStructures
{
	[Serializable()]
	public class Pattern
	{
		public static bool WeightSupport = false;
		public static bool WeightSize = false;

		public string EncodedPattern { get; set; }

		[field: NonSerialized()]
		private XElement xPattern;
		public XElement XPattern
		{
			get
			{
				if (this.xPattern == null)
				{
					this.xPattern = this.EncodedPattern.Parse();
				}

				return this.xPattern;
			}
		}

		public int Class { get; set; }

		protected double? distance;
		public double Distance
		{
			get
			{
				if (this.distance.HasValue)
				{
					return this.distance.Value;
				}
				else
				{
					throw new Exception("Niezainicjowana odległość!");
				}
			}
		}

		public double Support { get; set; }

		public double Confidence { get; set; }

		private double? size;
		public double Size
		{
			get
			{
				if (size.HasValue)
				{
					return size.Value;
				}
				else
				{
					return this.XPattern.DescendantNodesAndSelf().Count();
				}
			}
			set
			{
				this.size = value;
			}
		}

		public double CalculateDistance(Document document)
		{
			this.distance = PTED.Dynamic(this.XPattern, document.XDocument) / (double)this.XPattern.DescendantsAndSelf().Count();
			return this.Distance;
		}

		public static void NormalizeWeights(IEnumerable<Pattern> patterns)
		{
			double maxSup = 0;
			double maxSize = 0;

			foreach (var pattern in patterns)
			{
				maxSup = pattern.Support > maxSup ? pattern.Support : maxSup;
				maxSize = pattern.Size > maxSize ? pattern.Size : maxSize;
			}

			foreach (var pattern in patterns)
			{
				pattern.Support = pattern.Support / maxSup;
				pattern.Size = pattern.Size / maxSize;
			}
		}

		public double GetDistance()
		{
			return this.Distance;
		}

		public string GetClass()
		{
			return this.Class.ToString();
		}

		public double GetWeight()
		{
			double totalWeight = 1;

			if (Pattern.WeightSupport)
			{
				totalWeight *= (double)this.Support;
			}
			if (Pattern.WeightSize)
			{
				totalWeight *= (double)this.Size;
			}

			return totalWeight;
		
		}

		private int[,] ParsePattern(string[] patSplit)
		{
			int[,] pat = new int[patSplit.Count(x => int.Parse(x) >= 0), 2];

			int patLev = 0;
			int patPos = 0;
			for (int i = 0; i < patSplit.Length; i++)
			{
				int patVal = int.Parse(patSplit[i]);
				if (patVal < 0)
				{
					patLev--;
				}
				else
				{
					pat[patPos, 0] = patVal;
					pat[patPos, 1] = patLev;
					patLev++;
					patPos++;
				}
			}
			return pat;
		}
	}
}
