﻿using System;
using System.Collections.Generic;
using XmlClassifier.DataStructures;

namespace XmlStreamClassifier.ConsoleApplication
{
    /// <summary>
    /// Class for generating an Arff file from a dataset
    /// </summary>
	class ArffGenerator
	{
        ClassModel Model { get; set; }
		private List<Document> documentBuffer;

		private double minsup;
		private int treeMiningBatchSize;

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="minsup">Minimal support of frequent trees for a single batch in a single class</param>
		/// <param name="batchSize">Size of a single batch. Depends on <paramref name=synchronizedUpdate""/> parameter.
		/// For synchronized update means the total number of documents in all classes.
		/// For unsynchronized update means the number of documents in a single class.</param>
		/// <param name="synchronizedUpdate">Decides whether all classes are updated synchronously or not.</param>
        public ArffGenerator(double minsup, int batchSize, bool synchronizedUpdate, bool weightSize, bool weightSupport)
		{
			Pattern.WeightSize = weightSize;
			Pattern.WeightSupport = weightSupport;

			this.Model = new ClassModel(-1, minsup, batchSize, false, batchSize);
			this.documentBuffer = new List<Document>();

			this.minsup = minsup;
			this.treeMiningBatchSize = batchSize;
		}

        /// <summary>
        /// Returns an arff-compatible tuple for a given document
        /// </summary>
        /// <param name="document">Document to generate tuple from</param>
        /// <param name="numOfAttrs">Number of attributes in a single tuple</param>
        /// <returns>An arff-compatible tuple of the passed document</returns>
		public string GetTuple(Document document, int numOfAttrs)
		{
            string tuple = string.Empty;
            Model.AddDocument(document);
            Model.UpdateModel(false);

            for (int pattern = 0; pattern < this.Model.Model.Count; pattern++)
            {
                if (this.Model.Model[pattern].CalculateDistance(document) == 0)
                {
                    tuple += "1,";
                }
                else
                {
                    tuple += "0,";
                }
            }

            if (this.Model.Model.Count > numOfAttrs)
            {
                throw new Exception("More patterns than expected attributes!!!");
            }

            if (this.Model.Model.Count < numOfAttrs)
            {
                for (int i = 0; i < numOfAttrs-this.Model.Model.Count; i++)
                {
                    tuple += "0,";
                }
            }

            tuple += document.ExpectedClass.ToString();

            return tuple;
		}
	}
}
