﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using AdaTreeNat;
using XmlClassifier.Algorithms;
using XmlClassifier.DataStructures;

namespace XmlStreamClassifier.ConsoleApplication
{
	/// <summary>
	/// A classification model for a single class
	/// </summary>
	class ClassModel
	{
		/// <summary>
		/// Class of the model
		/// </summary>
		public int Class { get; set; }

		/// <summary>
		/// Actual model - a set of patterns.
		/// </summary>
		public List<Pattern> Model { get; set; }

		/// <summary>
		/// Bufor dokumentów
		/// </summary>
		public List<Document> Buffer { get; private set; }

		private AdaTreeNatWrapper treeMiner;
		private double minsup;
		private int batchSize;

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="_class">Class of the model</param>
		/// <param name="minsup">Minimal support of frequent patterns in the buffer</param>
		/// <param name="batchSize">Size of the buffer</param>
		public ClassModel(int _class, double minsup, int batchSize, bool isWindow, int windowSize)
		{
			this.Model = new List<Pattern>();
			this.Class = _class;
			this.minsup = minsup;
			this.batchSize = batchSize;
			this.treeMiner = new AdaTreeNatWrapper(true, false, true, isWindow, false, false, false, false, this.batchSize, isWindow ? windowSize : batchSize, 10000, this.minsup, true);
			this.Buffer = new List<Document>();
		}

		/// <summary>
		/// Adds new document to the buffer
		/// </summary>
		/// <param name="document">A new document</param>
		public void AddDocument(Document document)
		{
			this.Buffer.Add(document);
		}

		/// <summary>
		/// Updates the model - replaces old patterns with new ones.
		/// </summary>
		/// <param name="synchronizedUpdate">Decides the moment of model update. <c>true</c> forces the update while <c>false</c> waits untill the buffer is full.</param>
		public bool UpdateModel(bool synchronizedUpdate)
		{
			if (this.Buffer.Count > 0 && (synchronizedUpdate || this.Buffer.Count >= batchSize))
			{
				var newModel = this.treeMiner.GetPatterns(this.Buffer.Select(x => x.Content).ToList(), this.Class);
				this.Buffer.Clear();
				if (newModel.Any())
				{
					this.Model = newModel;
					return true;
				}
			}

			return false;
		}
	}
}
